# frozen_string_literal: true

require 'fileutils'

module Lexy
  class SpecFinder
    def call(path)
      stack = build_dirs_stack(path)
      spec_path = build_spec_path(path, stack)
      const = extract_constant_name(path)
      ensure_folder_created(spec_path)
      create_spec_file(spec_path, const)
    end

    def build_dirs_stack(path)
      current_path = path.split('/')
      dirs_stack = []
      dir_found = false

      until dir_found
        if specs_dir?(File.join(*current_path))
          dir_found = true
        else
          dirs_stack << current_path[-1]

          current_path = current_path[0...-1]

          raise 'No specs folder in given tree' if current_path.length.zero?
        end
      end

      dirs_stack + ['spec']
    end

    def specs_dir?(dir)
      Dir.glob('spec', base: dir).any?
    end

    def build_spec_path(path, dirs_stack)
      base_path = path.split('/') - dirs_stack
      spec_name = dirs_stack[0].split('.')[0] + '_spec.rb'
      spec_path = dirs_stack[1..-1].reverse

      base_path + spec_path + [spec_name]
    end

    def extract_constant_name(path)
      file_contents = File.read(path)

      ast = Parser::CurrentRuby.parse(file_contents)

      const = process_scope(ast)
      raise 'No constants definitions' if const.nil?

      const
    end

    def process_scope(ast, const = [])
      return unless ast.is_a?(Parser::AST::Node)

      names = []
      local_const = const.dup

      if %i[module class].include?(ast.type)
        name = node_name(ast)

        local_const << name
        names << local_const
      end

      nodes_names = ast.to_a.flat_map { |node| process_scope(node, local_const) }.reject(&:nil?)
      names.push(*nodes_names)

      return if names.empty?

      names
    end

    def ensure_folder_created(path)
      FileUtils.mkdir_p(File.join(path[0...-1]))
    end

    def create_spec_file(path, const)
      path = File.join(path)

      return path if File.exist?(path)

      template = <<~SPEC
        # frozen_string_literal: true

        RSpec.describe #{const.join('::')} do
          subject { described_class.new }
        end
      SPEC

      File.open(path, 'a+') { |file| file.write(template) }
      path
    end

    def node_name(ast)
      ast.to_a.first.to_a[1]
    end
  end
end
